"use strict";

var Promise = require("bluebird");

var logger = require('../logger');
var Functions = require('../Functions');

var AbstractCustomMap = require('./AbstractCustomMap');
var AbstractDbOverlay = require('./overlay/AbstractDbOverlay');
var AliasMarker = require('./marker/AliasMarker');
var CommentDialog = require('../gui/CommentDialog');
var ControlType = require('./ControlType');
var CustomMapOptions = require('./CustomMapOptions');
var IdentifiedElement = require('./data/IdentifiedElement');
var PointMarker = require('./marker/PointMarker');
var ReactionMarker = require('./marker/ReactionMarker');
var ReactionOverlay = require('./overlay/ReactionOverlay');
var ReferenceGenome = require('./data/ReferenceGenome');
var Submap = require('./Submap');
var TouchMap = require('./TouchMap');

/**
 * Default constructor.
 * 
 * @param globalMap
 *          google.maps.Map object representing the map
 * @param configuration
 *          Configuration object representing our data in the map
 * @param bigButtons
 *          boolean value determining if the buttons on the map should be big,
 *          and if the map is run on the touch interface
 * @param hideDiv
 * 
 */
function CustomMap(options) {
  this._controls = [];
  
  if (!(options instanceof CustomMapOptions)) {
    options = new CustomMapOptions(options);
  }
  AbstractCustomMap.call(this, options.getProject().getModel(), options);
  this.setProject(options.getProject());

  this.selectedLayouts = [];

  this.customizeGoogleMapView(options.getMapDiv());

  this.createMapChangedCallbacks();

  this.overlayCollections = [];

  // which submap is active (where user made interaction for the last time)
  this._activeSubmapId = null;

  this.initialized = true;

  // list of reference genomes
  this._referenceGenome = [];
  
  if (options.isCustomTouchInterface()) {
    this._touchInterface = new TouchMap(this);
  }

  this.createSubmaps();

  this._dialogs = [];
}

CustomMap.prototype = Object.create(AbstractCustomMap.prototype);
CustomMap.prototype.constructor = CustomMap;

CustomMap.prototype.createSubmaps = function() {
  this.submaps = [];
  for (var i = 0; i < this.getModel().getSubmodels().length; i++) {
    this.submaps.push(new Submap(this, this.getModel().getSubmodels()[i]));
  }
};

CustomMap.prototype.createLogo = function() {
  var logoControlDiv2 = document.createElement('DIV');
  var logo2 = document.createElement('IMG');
  ServerConnector.getLogoImg().then(function(url){
    if (!/^(f|ht)tps?:\/\//i.test(url)) {
      url = GuiConnector.getImgPrefix() + url;
    }
    logo2.src = url;
  });
  logo2.style.cursor = 'pointer';
  logo2.style.width = "80px";
  logo2.id = ControlType.LOGO_2_IMG;
  this.addControl(logo2);
  
  logoControlDiv2.appendChild(logo2);
  google.maps.event.addDomListener(logo2, 'click', function() {
    return ServerConnector.getLogoLink().then(function(url){
      var win = window.open(url, '_blank');
      win.focus();
    }).catch(GuiConnector.alert);
  });
  logoControlDiv2.index = 0; // used for ordering
  this.getGoogleMap().controls[google.maps.ControlPosition.LEFT_BOTTOM].push(logoControlDiv2);

  var logoControlDiv = document.createElement('DIV');
  logoControlDiv.style.padding = '5px';

  var logo = document.createElement('IMG');
  logo.src = GuiConnector.getImgPrefix() + GuiConnector.getLcsbLogoImg(this.bigButtons);
  logo.style.cursor = 'pointer';
  logo.id = ControlType.LOGO_IMG;
  this.addControl(logo);

  logoControlDiv.appendChild(logo);
  google.maps.event.addDomListener(logo, 'click', function() {
    var win = window.open('http://wwwen.uni.lu/lcsb/', '_blank');
    win.focus();
  });

  logoControlDiv.index = 1; // used for ordering
  this.getGoogleMap().controls[google.maps.ControlPosition.RIGHT_BOTTOM].push(logoControlDiv);
};

CustomMap.prototype.clearOverlays = function() {
  for ( var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      if (collection.constructor.name==="OverlayCollection") {
        this.clearOverlayCollection(collection);
      }else {
        collection.clear();
      }
    }
  }
};

/**
 * Removes all markers from {@link OverlayCollection}.
 * 
 * @param collection
 *          {@link OverlayCollection} from which all markers should be removed
 */
CustomMap.prototype.clearOverlayCollection = function(collection) {
  logger.debug("Clear collection: " + collection.name);
  for ( var aliasKey in collection.aliasMarkers) {
    if (collection.aliasMarkers.hasOwnProperty(aliasKey) && collection.aliasMarkers[aliasKey] !== undefined
        && collection.aliasMarkers[aliasKey] !== null) {
      collection.aliasMarkers[aliasKey].setMap(null);
    }
  }

  for ( var pointKey in collection.pointMarkers) {
    if (collection.pointMarkers.hasOwnProperty(pointKey) && collection.pointMarkers[pointKey] !== undefined
        && collection.pointMarkers[pointKey] !== null) {
      collection.pointMarkers[pointKey].setMap(null);
    }
  }

  for ( var key in collection.reactionMarkers) {
    if (collection.reactionMarkers.hasOwnProperty(key) && collection.reactionMarkers[key] !== undefined
        && collection.reactionMarkers[key] !== null) {
      collection.reactionMarkers[key].setMap(null);
    }
  }

  collection.aliasMarkers = [];
  collection.pointMarkers = [];
  collection.reactionMarkers = [];
};

/**
 * Updates data about visualized markers in {@link OverlayCollection}.
 * 
 * @param overlayCollection
 *          {@link OverlayCollection} with new data to visualize
 * @param fitBounds
 *          <code>true</code> id the map should fit bounds to the new elements
 *          after update, <code>false</code> otherwise
 */
CustomMap.prototype.updateOverlayCollection = function(overlayCollection, fitBounds) {
  this.clearOverlayCollection(overlayCollection);
  return this.renderOverlayCollection({overlayCollection:overlayCollection, fitBounds:fitBounds});
};

/**
 * This method open layout by a given layout identifier (string starting with
 * 'cv' prefix) in a map and all submaps.
 * 
 * @param identifier
 *          identifier of the layout to present
 */
CustomMap.prototype.openLayout = function(identifier) {
  var self = this;
  
  logger.debug("Opening layout: " + identifier);

  this.getGoogleMap().setMapTypeId(identifier);

  var index = null;
  for (var j = 0; j < this.getLayouts().length; j++) {
    var layout = this.getLayouts()[j];
    if ('cv' + layout.getId() === identifier) {
      index = j;
    }
  }
  if (index === null) {
    logger.warn("Invalid layout identifier: " + identifier);
  }
  var submaps = self.getSubmaps();
  for (var i = 0; i < submaps.length; i++) {
    var submap = submaps[i];
    submap.openLayout('cv' + submap.getModel().getLayouts()[index].getId());
  }
};

/**
 * This method open layout by a given database identifier.
 * 
 * @param identifier
 *          identifier of the layout to present
 */
CustomMap.prototype.openLayoutById = function(identifier) {
  identifier = parseInt(identifier);
  logger.debug("Opening layout: " + identifier);
  var index = null;
  for (var i = 0; i < this.getLayouts().length; i++) {
    var layout = this.getLayouts()[i];
    if (layout.getId() === identifier) {
      index = 'cv' + identifier;
    }
  }

  // if layout doesn't exist print error
  if (index === null) {
    throw new Error("You have no privileges for selected layout");
  } else {
    this.openLayout(index);
  }
};

CustomMap.prototype.openLayoutByName = function(name) {
  logger.debug("Opening layout: " + name);
  var index = null;
  var layouts = this.getLayouts();
  for (var i = 0; i < layouts.length; i++) {
    var layout = layouts[i];
    if (layout.getName() === name) {
      index = 'cv' + layout.getId();
    }
  }
  // if layout doesn't exist print error
  if (index === null) {
    alert("You have no privileges for selected layout");
  } else {
    this.openLayout(index);
  }
};

CustomMap.prototype.registerSource = function(overlayCollection) {
  var self = this;

  this.overlayCollections[overlayCollection.name] = overlayCollection;

  overlayCollection.aliasMarkers = [];
  overlayCollection.pointMarkers = [];
  overlayCollection.reactionMarkers = [];

  if (overlayCollection.getName()==="search") {
    overlayCollection.addListener("onSearch", function() {
      return self.updateOverlayCollection(overlayCollection, false);
    });
  }
  if (overlayCollection.getName() === "drug" || overlayCollection.getName() === "chemical" ||overlayCollection.getName() === "mirna" ) {
    overlayCollection.addListener("onSearch", function() {
      return self.updateOverlayCollection(overlayCollection, false);
    });
    
    overlayCollection.addListener("onTargetVisibilityChange", function() {
      return self.updateOverlayCollection(overlayCollection, false);
    });
  }

};

CustomMap.prototype.refreshComments = function() {
  var self = this;
  
  var collection = self.getOverlayByName("comment");
  if (collection!==undefined) {
    self.clearOverlayCollection(collection);
    if (ServerConnector.getSessionData(self.getProject()).getShowComments()) {
      return collection.refresh().then(function(){
        return self.renderOverlayCollection({overlayCollection:collection});        
      });
    } else {
      return collection.clear();
    }
  }
  throw new Error("comment OverlayCollection not found");
};

CustomMap.prototype.turnOnOffDrawing = function() {
  var model = this.getSubmodelById(this.getActiveSubmapId());
  if (model !== undefined) {
    model._turnOnOffDrawing();
  } else {
    throw "Cannot find submodel with id: " + this.getActiveSubmapId();
  }
};

CustomMap.prototype.clearData = function() {
  this.clearOverlays();
};

CustomMap.prototype.refreshMarkers = function() {
  logger.debug("Refresh Markers: ");
  for ( var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var collection = this.overlayCollections[overlayName];
      this.refreshOverlayMarkers(collection);
    }
  }
};

CustomMap.prototype.refreshOverlayMarkers = function(overlay) {
  logger.debug("Refresh overlay: " + overlay.name);

  for ( var aliasKey in overlay.aliasMarkers) {
    if (overlay.aliasMarkers.hasOwnProperty(aliasKey) && overlay.aliasMarkers[aliasKey] !== undefined
        && overlay.aliasMarkers[aliasKey] !== null) {
      var alias = overlay.aliasMarkers[aliasKey];
      if (!this.isMarkerOptimization()) {
        alias.hide();
        alias.show();
      }
    }
  }

  for ( var key in overlay.pointMarkers) {
    if (overlay.pointMarkers.hasOwnProperty(key) && overlay.pointMarkers[key] !== undefined
        && overlay.pointMarkers[key] !== null) {
      var pointMarker = overlay.pointMarkers[key];
      // we don't need to update this markers because thet data about
      // visualization is
      // already there
      // alias.update();
      if (!this.isMarkerOptimization()) {
        pointMarker.hide();
        pointMarker.show();
      }
    }
  }

  for ( var reactionKey in overlay.reactionMarkers) {
    if (overlay.reactionMarkers.hasOwnProperty(reactionKey) && overlay.reactionMarkers[reactionKey] !== undefined
        && overlay.reactionMarkers[reactionKey] !== null) {
      var reactionOverlay = overlay.reactionMarkers[reactionKey];
      if (!reactionOverlay.isShown()) {
        reactionOverlay.show();
      }
      if (!this.isMarkerOptimization()) {
        reactionOverlay.hide();
        reactionOverlay.show();
      }
    }
  }
};

CustomMap.prototype.getSubmapDialogDiv = function(id) {
  var dialogDiv = this._dialogs[id];

  if (dialogDiv === undefined) {
    dialogDiv = document.createElement("div");
    dialogDiv.setAttribute("name", "dialog-" + id);

    this._dialogs[id] = dialogDiv;
    
    $(dialogDiv).dialog({autoOpen: false});
  }
  return dialogDiv;
};

CustomMap.prototype.openSubmodel = function(id) {
  var self = this;
  var submap = self.getSubmodelById(id);
  if (submap === null) {
    throw new Error("Unknown submap for id: " + id);
  }
  if (submap!==this) {

    var dialogDiv = self.getSubmapDialogDiv(id);

    submap.open(dialogDiv);

    // we have to perform it on top map, because on submaps id is different
    this.openLayout(this.getGoogleMap().getMapTypeId());

    // now we have to visualize layouts
    var layouts = [];

    // get list of layouts
    for ( var key in this.selectedLayouts) {
      if (this.selectedLayouts.hasOwnProperty(key) && this.selectedLayouts[key] === true) {
        layouts.push(key);
      }
    }

    // show layouts that should be visualized (resize or show them)
    for (var i = 0; i < layouts.length; i++) {
      var layoutId = layouts[i];
      submap._showSelectedLayout(layoutId, i, layouts.length);
    }
  }

};

CustomMap.prototype.customizeGoogleMapView = function(div) {
  var mapOptions = this.creatMapOptions();
  this.setGoogleMap(new google.maps.Map(div, mapOptions));
  this.setupLayouts();

  this.registerMapClickEvents();

  this.createLogo();
  // this.createMapVersion();
  google.maps.event.trigger(this.getGoogleMap(), 'resize');
  google.maps.event.trigger(this.getGoogleMap(), 'maptypeid_changed');
  google.maps.event.trigger(this.getGoogleMap(), 'projection_changed');

  // center map and zoom in to fit into browser window if there is no
  // information about coordinates in the session
  if (ServerConnector.getSessionData(this.getProject()).getCenter(this.getModel())===undefined) {
    var bounds = new google.maps.LatLngBounds();
    bounds.extend(this.getTopLeftLatLng());
    bounds.extend(this.getBottomRightLatLng());

    this.getGoogleMap().fitBounds(bounds);
  }
};

CustomMap.prototype.setCenter = function(mapIdentifier, coordinates) {
  if (this.getModel().getId() === mapIdentifier) {
    this.getGoogleMap().setCenter(coordinates);
  } else {
    GuiConnector.openDialog(mapIdentifier);
    for (var i = 0; i < this.submaps.length; i++) {
      if (this.submaps[i].getId() === mapIdentifier) {
        if (coordinates instanceof google.maps.Point) {
          coordinates = this.submaps[i].fromPointToLatLng(coordinates);
        }
        this.submaps[i].getGoogleMap().setCenter(coordinates);
      }
    }
  }
};

CustomMap.prototype.setZoom = function(mapIdentifier, zoom) {
  if (this.getModel().getId() === mapIdentifier) {
    this.getGoogleMap().setZoom(zoom);
  } else {
    GuiConnector.openDialog(mapIdentifier);
    for (var i = 0; i < this.submaps.length; i++) {
      if (this.submaps[i].getId() === mapIdentifier) {
        this.submaps[i].getGoogleMap().setZoom(zoom);
      }
    }
  }
};

/**
 * Creates listeners for google.maps.Map object that will actualize the data in
 * user session.
 */
CustomMap.prototype.createMapChangedCallbacks = function() {
  var customMapSelf = this;
  var sessionData = ServerConnector.getSessionData(customMapSelf.getProject()); 
  // listener for changing zoom level
  google.maps.event.addListener(this.getGoogleMap(), 'zoom_changed', function() {
    sessionData.setZoomLevel(customMapSelf.getModel(),customMapSelf.getGoogleMap().getZoom());
  });

  // if we have zoom level data stored in session then restore it
  var level = ServerConnector.getSessionData(customMapSelf.getProject()).getZoomLevel(this.getModel());
  if (parseInt(level) > 0) {
    level = parseInt(level);
    this.getGoogleMap().setZoom(level);
  } else {
    sessionData.setZoomLevel(customMapSelf.getModel(),customMapSelf.getGoogleMap().getZoom());
  }

  // listener for changing location of the map (moving left/reght/top/bottom
  google.maps.event.addListener(this.getGoogleMap(), 'center_changed', function() {
    var coord = customMapSelf.getGoogleMap().getCenter();
    var point = customMapSelf.fromLatLngToPoint(coord);
    sessionData.setCenter(customMapSelf.getModel(),point);
  });

  // listener for changing type of layout
  google.maps.event.addListener(this.getGoogleMap(), 'maptypeid_changed', function() {
    sessionData.setSelectedBackgroundOverlay(customMapSelf.getGoogleMap().getMapTypeId());
  });

  // if we have type of layout stored in the session then restore it
  var mapType = sessionData.getSelectedBackgroundOverlay();
  if (mapType !== undefined) {
    this.openLayout(mapType);
  }
};

/**
 * Returns submodel (or this model) by identfier of the model.
 * 
 * @param identifier
 *          identifier of the submodel
 * @returns submodel (or this model) with given identfier of the model
 */
CustomMap.prototype.getSubmodelById = function(identifier) {
  identifier = parseInt(identifier);
  if (this.getId() === identifier) {
    return this;
  }
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].getId() === identifier) {
      return this.submaps[i];
    }
  }
  logger.warn("Cannot find submodel with id: "+identifier);
  return null;
};

CustomMap.prototype.removeSelection = function() {
  var model = this.getSubmodelById(this.getActiveSubmapId());
  if (model !== null) {
    model._removeSelection();
  } else {
    throw "Cannot find submodel with id: " + this.getActiveSubmapId();
  }
};

CustomMap.prototype.showModel = function(id, point, zoomLevel) {
  if (point !== undefined) {
    this.setCenter(id, point);
  } else {
    logger.warn("Center point undefined...");
  }
  if (zoomLevel !== undefined) {
    this.setZoom(id, zoomLevel);
  } else {
    logger.warn("Zoom level undefined...");
  }
};

/**
 * Adds information about aliases visible in the specific layout.
 * 
 * @param layoutId
 *          identifier of the layout
 * 
 * @param jsonAliases
 *          list of aliases in json format
 * 
 */
CustomMap.prototype.addAliasesForLayout = function(layoutId, jsonAliases) {
  logger.debug("Adding aliases for layout: " + layoutId);

  // init layout data
  if (this.getModel().getLayoutDataById(layoutId) === undefined) {
    this.getModel().initLayoutData(layoutId);
    for (var j = 0; j < this.submaps.length; j++) {
      this.submaps[j].getModel().initLayoutData(layoutId);
    }
  }

  var aliases = JSON.parse(jsonAliases);
  for (var i = 0; i < aliases.length; i++) {
    var alias = aliases[i];
    var model = this.getSubmodelById(alias.modelId);
    if (model !== null) {
      model.getModel().addAliasForLayout(layoutId, alias);
    } else {
      logger.warn("Unknown model: " + alias.modelId);
    }
  }

  this.retrieveMissingAliases();
};

/**
 * Adds information about aliases.
 * 
 * @param jsonAliases
 *          list of aliases in json format
 * 
 */
CustomMap.prototype.addAliases = function(aliases) {
  for (var i = 0; i < aliases.length; i++) {
    var alias = aliases[i];
    var model = this.getSubmodelById(alias.modelId);
    if (model !== undefined) {
      model.addAlias(alias);
    } else {
      logger.warn("Unknown model: " + alias.modelId);
    }
  }
  this.callListeners("onAddAliases");
};

/**
 * This function will ask server for aliases that should be visualized but the
 * data is still missing on the client side.
 */
CustomMap.prototype.retrieveMissingAliases = function() {
  var ids = [];
  var missing = this.getModel().getMissingAliasIds();
  for (var k = 0; k < missing.length; k++) {
    ids.push([ this.getId(), missing[k] ]);
  }
  for (var i = 0; i < this.submaps.length; i++) {
    missing = this.submaps[i].getModel().getMissingAliasIds();
    for (var j = 0; j < missing.length; j++) {
      ids.push([ this.submaps[i].getId(), missing[j] ]);
    }
  }
  if (ids.length > 0) {
    // load data from server about missing aliases
    ServerConnector.retreiveLightAliases(ids);
  }
  if (!ServerConnector.isWaitingForData()) {
    // if we already have everything then just refresh data to be visualized
    this.refreshSelectedLayouts();
    // and close "loading" dialog
    GuiConnector.closeLoadingDialog();
  }
};

/**
 * Adds layout to be visualized.
 * 
 * @param identifier
 *          identifier of the layout that should be included in visualization
 */
CustomMap.prototype.addSelectedLayout = function(identifier) {
  var self = this;
  return new Promise(function(resolve, reject) {
    logger.debug("Selecting layout: " + identifier);

    if (self.selectedLayouts[identifier] === true) {
      logger.warn("Layout " + identifier + " already selected");
      resolve();
    } else {
      self.selectedLayouts[identifier] = true;
      ServerConnector.getSessionData(self.getProject()).setVisibleOverlays(self.getSelectedLayouts());
      return self.getModel().initLayoutDataByLayoutId(identifier).then(function() {
        return self.refreshSelectedLayouts();
      }).then(function() {
        resolve();
      }).catch(function(exception){
        reject(exception);
      });
    }
  });
};

/**
 * Returns list of layouts that are selected and visualized by javascript.
 * 
 * @return array with a list of selected layotus
 * 
 */
CustomMap.prototype.getSelectedLayouts = function() {
  var layouts = [];

  // get list of layouts
  for ( var key in this.selectedLayouts) {
    if (this.selectedLayouts.hasOwnProperty(key) && this.selectedLayouts[key] === true) {
      layouts.push(key);
    }
  }
  return layouts;
};

/**
 * Removes layout from visualization.
 * 
 * @param identifier
 *          identifier of layout to remove
 * 
 */
CustomMap.prototype.removeSelectedLayout = function(identifier) {
  logger.debug("Removing layout: " + identifier);

  if (this.selectedLayouts[identifier] !== true) {
    logger.warn("Layout " + identifier + " is not selected");
  } else {
    this.selectedLayouts[identifier] = false;
    this.refreshSelectedLayouts();
    ServerConnector.getSessionData(this.getProject()).setVisibleOverlays(this.getSelectedLayouts());
  }
};

/**
 * Refresh visualization of selected layouts.
 */
CustomMap.prototype.refreshSelectedLayouts = function() {
  logger.debug("Refreshing layouts");
  var self = this;
  var layouts = self.getSelectedLayouts();
  // show layouts that should be visualized (resize or show them)
  var promises = [];
  for (var i = 0; i < layouts.length; i++) {
    var layoutId = layouts[i];
    if (self.layoutContainsOverlays(layoutId)) {
      // resize element on the map
      promises.push(self.resizeSelectedLayout(layoutId, i, layouts.length));
    } else {
      promises.push(self.showSelectedLayout(layoutId, i, layouts.length));
    }
  }
  return Promise.all(promises).then(function() {
    // remove layouts that were
    for ( var key in self.selectedLayoutOverlays) {
      if (!self.selectedLayouts.hasOwnProperty(key) || self.selectedLayouts[key] === false) {
        if (self.layoutContainsOverlays(key)) {
          self.hideSelectedLayout(key);
        }
      }
    }
    self.refreshInfoWindows();    

  });
};

/**
 * Hides layout from the map and all submaps
 * 
 * @param layoutId
 *          identifier of a layout to hide
 */
CustomMap.prototype.hideSelectedLayout = function(layoutId) {
  this._hideSelectedLayout(layoutId);
  for (var i = 0; i < this.submaps.length; i++) {
    this.submaps[i]._hideSelectedLayout(layoutId);
  }
};

/**
 * Resize(refresh) layout on the map and all submaps. Resizing should be called
 * when number of layouts to visualize change.
 * 
 * @param layoutId
 *          identifier of layout to refresh
 * @param index
 *          position of the layout in list of layouts that we visualize
 * @param length
 *          number of layouts that we currently visualize
 */
CustomMap.prototype.resizeSelectedLayout = function(layoutId, index, length) {
  logger.debug("Resize layout: " + layoutId);
  var promises = [];
  promises.push(this._resizeSelectedLayout(layoutId, index, length));
  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._resizeSelectedLayout(layoutId, index, length));
  }
  return Promise.all(promises);
};

/**
 * Show layout on the map and all submaps.
 * 
 * @param layoutId
 *          identifier of layout to show
 * @param index
 *          position of the layout in list of layouts that we visualize
 * @param length
 *          number of layouts that we currently visualize
 */
CustomMap.prototype.showSelectedLayout = function(layoutId, index, length) {
  logger.debug("Resize layout: " + layoutId);
  var promises = [];
  promises.push(this._showSelectedLayout(layoutId, index, length));

  for (var i = 0; i < this.submaps.length; i++) {
    promises.push(this.submaps[i]._showSelectedLayout(layoutId, index, length));
  }
  return Promise.all(promises);
};

/**
 * Adds information about reactions visible in the specific layout.
 * 
 * @param layoutId
 *          identifier of the layout
 * 
 * @param jsonAliases
 *          list of reactions in json format
 * 
 */
CustomMap.prototype.addReactionsForLayout = function(layoutId, jsonReactions) {
  logger.debug("Adding reactions for layout: " + layoutId);
  var reactions = JSON.parse(jsonReactions);
  for (var i = 0; i < reactions.length; i++) {
    var reaction = reactions[i];
    var model = this.getSubmodelById(reaction.modelId);
    if (model !== undefined) {
      model.getModel().addReactionForLayout(layoutId, reaction);
    } else {
      logger.warn("Unknown model: " + reaction.modelId);
    }
  }
  this.retrieveMissingReactions();
};

/**
 * Adds information about reactions.
 * 
 * @param jsonAliases
 *          list of reactions in json format
 * 
 */
CustomMap.prototype.addReactions = function(jsonReactions) {
  var reactions = JSON.parse(jsonReactions);
  for (var i = 0; i < reactions.length; i++) {
    var reaction = reactions[i];
    var model = this.getSubmodelById(reaction.modelId);
    if (model !== undefined) {
      model.addReaction(reaction);
    } else {
      logger.warn("Unknown model: " + reaction.modelId);
    }
  }
  this.callListeners("onAddReactions");
};

/**
 * This function will ask server for reactions that should be visualized but the
 * data is still missing on the client side.
 */
CustomMap.prototype.retrieveMissingReactions = function() {
  var ids = [];
  var missing = this.getModel().getMissingReactionIds();
  for (var k = 0; k < missing.length; k++) {
    ids.push([ this.getId(), missing[k] ]);
  }
  for (var i = 0; i < this.submaps.length; i++) {
    missing = this.submaps[i].getModel().getMissingReactionIds();
    for (var j = 0; j < missing.length; j++) {
      ids.push([ this.submaps[i].getId(), missing[j] ]);
    }
  }
  if (ids.length > 0) {
    // load data from server about missing reactions
    ServerConnector.retreiveLightReactions(ids);
  }
  if (!ServerConnector.isWaitingForData()) {
    // if we already have everything then just refresh data to be visualized
    this.refreshSelectedLayouts();
    // and close "loading" dialog (if opened)
    GuiConnector.closeLoadingDialog();
  }
};

/**
 * This method checks if the layout contains any overlays (like AliasOverlay or
 * ReactionOverlay) that is currently visible on the map.
 * 
 * @param layoutId
 *          identifier of the layout
 * @returns {Boolean}: <code>true</code> if the layout contains overlays to
 *          visualize, <code>false</code> otherwise
 */
CustomMap.prototype.layoutContainsOverlays = function(layoutId) {

  // first, check top map
  if (this.selectedLayoutOverlays.hasOwnProperty(layoutId) && this.selectedLayoutOverlays[layoutId].length > 0) {
    return true;
  }

  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    if (this.submaps[i].initialized) {
      if (this.submaps[i].selectedLayoutOverlays.hasOwnProperty(layoutId)
          && this.submaps[i].selectedLayoutOverlays[layoutId].length > 0) {
        return true;
      }
    }
  }
  return false;
};

/**
 * Refresh content of all {@link AliasInfoWindow} in this map and all submaps.
 */
CustomMap.prototype.refreshInfoWindows = function() {
  this._refreshInfoWindows();
  // now check all submaps
  for (var i = 0; i < this.submaps.length; i++) {
    this.submaps[i]._refreshInfoWindows();
  }
};

/**
 * Opens {@link AliasInfoWindow} for an {@link Alias} in a given model/submodel.
 * 
 * @param aliasId
 *          identifier of {@link Alias}
 * @param modelId
 *          identifier of {@link AbstractCustomMap}
 */
CustomMap.prototype.openInfoWindowForAlias = function(aliasId, modelId) {
  logger.debug("Opening info window for alias: " + aliasId + ", model: " + modelId);
  var model = this.getSubmodelById(modelId);
  return model.getModel().getAliasById(aliasId, true).then(function() {
    model._openInfoWindowForAlias(aliasId);
  });
};

/**
 * Renders markers, lines, etc. for elements highlighted in OverlayCollection.
 * 
 * @param overlayCollection
 *          {@link OverlayCollection} to be processed
 * @param fitBounds
 *          <code>true</code> if the borders should fit bounds after creating
 *          all elements
 */
CustomMap.prototype.renderOverlayCollection = function(overlayCollection, fitBounds) {
  var self = this;
  if (overlayCollection instanceof AbstractDbOverlay) {
    logger.warn("Deprecated method use: renderOverlayCollection(overlayCollection)");
  } else {
    fitBounds = overlayCollection.fitBounds;
    overlayCollection = overlayCollection.overlayCollection;
  }
  
  var elements;
  var boundsArray = [];
  boundsArray[self.getId()] = new google.maps.LatLngBounds();
  for (var j = 0; j < self.submaps.length; j++) {
    boundsArray[self.submaps[j].getId()] = new google.maps.LatLngBounds();
  }

  return overlayCollection.getIdentifiedElements().then(function(identifiedElements){
    elements = identifiedElements;

    var bounds;

    return Promise.each(
        elements,
        function(element) {
          var model = self.getSubmodelById(element.modelId);
          if (element.getType() === "ALIAS") {
            if (overlayCollection.aliasMarkers[element.getId()] !== undefined) {
              logger.warn("More than one marker in " + overlayCollection.name + " for alias " + element.getId()
                  + ". Skipping duplicates.");
              return null;
            } else {
              return model.getModel().getAliasById(element.getId()).then(function(aliasData) {
                var aliasMarker = new AliasMarker(element.getId(), element.icon, aliasData, model);
                overlayCollection.aliasMarkers[element.getId()] = aliasMarker;
                bounds = aliasMarker.getBounds();
                boundsArray[element.getModelId()].extend(bounds.getNorthEast());
                boundsArray[element.getModelId()].extend(bounds.getSouthWest());
                return aliasMarker;
              });
            }
          } else if (element.getType() === "REACTION") {
            return model.getModel().getReactionById(element.getId()).then(function(reactionData) {
              var marker = null;
              var icon = element.getIcon();

              if (icon === null || icon === undefined) {
                // this should happen when we visualize search data (there is
                // no marker, but only flat overlay of the reaction lines)
                //
                marker = new ReactionOverlay(null, reactionData, model, false, element.getId());
              } else {
                // when we have icon defined (for instance when it comes from
                // comment) then we don't want to have overlayed reaction lines
                // but icon marker
                marker = new ReactionMarker(element.getId(), icon, reactionData, model);
              }
              marker.show();
              overlayCollection.reactionMarkers[element.getId()] = marker;
              bounds = marker.getBounds();
              boundsArray[element.getModelId()].extend(bounds.getNorthEast());
              boundsArray[element.getModelId()].extend(bounds.getSouthWest());
              return marker;
            });
          } else if (element.getType() === "POINT") {
            var pointData = model.getModel().getPointDataByPoint(element.getPoint());
            if (overlayCollection.pointMarkers[pointData.getId()]  === undefined) {
              var pointMarker = new PointMarker(pointData, element.icon, model);
              overlayCollection.pointMarkers[pointData.getId()] = pointMarker;
              bounds = pointMarker.getBounds();
              boundsArray[element.getModelId()].extend(bounds.getNorthEast());
              boundsArray[element.getModelId()].extend(bounds.getSouthWest());
            }
            return overlayCollection.pointMarkers[pointData.getId()];
          } else {
            throw new Error("Unknown type of the element in overlay: " + element.type);
          }
        });
  }).then(function() {

    for (var i = 0; i < elements.length; i++) {
      var element = elements[i];
      var infoWindow = self.getInfoWindowForIdentifiedElement(element);
      if (infoWindow !== null && infoWindow !== undefined) {
        self.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray());
        self.updateInfoWindowForIdentifiedElement(element);
      }
    }

    if (elements.length > 0 && fitBounds) {
      for ( var mapId in boundsArray) {
        if (boundsArray.hasOwnProperty(mapId)) {
          var map = self.getSubmodelById(mapId).getGoogleMap();
          var bounds = boundsArray[mapId];
          if (map !== undefined && !bounds.isEmpty()) {
              map.fitBounds(bounds);
          }
        }
      }
    }

  });
};

/**
 * Opens {@link AbstractInfoWindow} for a marker.
 * 
 * @param marker
 *          {@link AbstractMarker} for which info window should be opened
 */
CustomMap.prototype.openInfoWindowForMarker = function(marker) {
  var modelId = marker.getCustomMap().getId();
  var markerId = marker.getId();
  var model = this.getSubmodelById(modelId);
  logger.debug("Opening info window for " + marker.constructor.name + ": " + markerId + ", model: " + modelId);
  var elementType = marker.getType();

  // open AliasInfoWindow in a right model
  model._openInfoWindowForMarker(marker);

  var infoWindow = model.returnInfoWindowForMarker(marker);

  var element = new IdentifiedElement({
    objectId : markerId,
    modelId : modelId,
    type : elementType
  });

  this.retrieveOverlayDetailDataForElement(element, infoWindow.getOverlayFullViewArray());

};

/**
 * Sends requestes to download detailed data about element in all
 * {@link OverlayCollection}.
 * 
 * @param element
 *          element for which we want to have detailed information
 */
CustomMap.prototype.retrieveOverlayDetailDataForElement = function(element, general) {
  var promises = [];
  if (general === undefined) {
    logger.warn("general param is undefined!");
    general = [];
  }
  for ( var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];

      var generalRequest = general[overlayName];
      if (generalRequest === undefined) {
        logger.warn("No information about general overlay request for overlay: ", overlayName);
        generalRequest = false;
      }
      generalRequest = generalRequest || !overlay.allowSearchById();

      promises.push(overlay.getDetailDataByIdentifiedElement(element, generalRequest));
    }
  }
  return Promise.all(promises);
};

/**
 * Updates info window identified by element given as a parameter.
 * 
 * @param identifiedElement
 *          element for which info window should be updated
 */
CustomMap.prototype.updateInfoWindowForIdentifiedElement = function(identifiedElement) {
  var infoWindow = this.getInfoWindowForIdentifiedElement(identifiedElement);
  if (infoWindow === null || infoWindow === undefined) {
    return;
  } else {
    infoWindow.update();
  }
};

/**
 * Returns data from all {@link OverlayCollection} for a given alias.
 * 
 * @param alias
 *          {@link Alias} for which overlay data will be returned
 * @param general
 *          if true then all elements will be returned, if false then only ones
 *          availble right now in the overlay
 * @returns data from all {@link OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForAlias = function(alias, general) {
  var identifiedElement = new IdentifiedElement(alias);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for a given reaction.
 * 
 * @param reaction
 *          {@link Reaction} for which overlay data will be returned
 * @param general
 *          if true then all elements will be returned, if false then only ones
 *          availble right now in the overlay
 * @returns data from all {@link OverlayCollection} for a given alias
 */
CustomMap.prototype.getOverlayDataForReaction = function(reaction, general) {
  var identifiedElement = new IdentifiedElement(reaction);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for a given {@link PointData}
 * 
 * @param point
 *          {@link PointData} for which overlay data will be returned
 * @returns data from all {@link OverlayCollection} for a given
 *          {@link PointData}
 */
CustomMap.prototype.getOverlayDataForPoint = function(point, general) {
  var identifiedElement = new IdentifiedElement(point);
  return this.getOverlayDataForIdentifiedElement(identifiedElement, general);
};

/**
 * Returns data from all {@link OverlayCollection} for element identified by the
 * parameter
 * 
 * @param identifiedElement
 *          {@link IdentifiedElement} for which overlay data will be returned
 * @returns data from all {@link OverlayCollection} for a given
 *          {@link IdentifiedElement}
 */
CustomMap.prototype.getOverlayDataForIdentifiedElement = function(identifiedElement, general) {
  if (general === undefined) {
    logger.warn("general parameter must be defined");
    general = [];
  }
  var promises = [];
  var overlays = [];
  for ( var overlayName in this.overlayCollections) {
    if (this.overlayCollections.hasOwnProperty(overlayName)) {
      var overlay = this.overlayCollections[overlayName];
      if (overlay.allowGeneralSearch() || overlay.allowSearchById()) {
        var generalFlag = general[overlay.getName()];
        if (generalFlag === undefined) {
          logger.warn("General flag for overlay: " + overlay.getName() + " is not defined, assuming false");
          generalFlag = false;
        }
        overlays.push(overlay);
        promises.push(overlay.getDetailDataByIdentifiedElement(identifiedElement, !overlay.allowSearchById() || generalFlag));
      }
    }
  }
  return new Promise(function(resolve,reject){
    Promise.all(promises).then(function(values){
      var result = [];
      for (var i=0;i<values.length;i++){
        result.push({overlay:overlays[i], data:values[i]});
      }
      resolve(result);
    }).catch(reject);
  });
};

CustomMap.prototype.getOverlayByName = function(name) {
  return this.overlayCollections[name];
};

/**
 * Returns {@link AbstractInfoWindow} for element identified by the parameter.
 * 
 * @param identifiedElement
 *          {@link IdentifiedElement} that determines for which element we want
 *          {@link AbstractInfoWindow}
 * @returns {@link AbstractInfoWindow} for element identified by the parameter
 */
CustomMap.prototype.getInfoWindowForIdentifiedElement = function(identifiedElement) {
  var model = this.getSubmodelById(identifiedElement.modelId);
  var infoWindow = null;
  if (identifiedElement.type === "ALIAS") {
    infoWindow = model.getAliasInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "POINT") {
    infoWindow = model.getPointInfoWindowById(identifiedElement.getId());
  } else if (identifiedElement.type === "REACTION") {
    infoWindow = model.getReactionInfoWindowById(identifiedElement.getId());
  } else {
    throw new Error("Unknown type of IdentifiedElement: ", identifiedElement);
  }
  return infoWindow;
};

CustomMap.prototype.getActiveSubmapId = function() {
  return this._activeSubmapId;
};

CustomMap.prototype.setActiveSubmapId = function(submapId) {
  this._activeSubmapId = submapId;
};

CustomMap.prototype.setActiveSubmapClickCoordinates = function(coordinates) {
  if (!(coordinates instanceof google.maps.Point)) {
    throw new Error("Coordinates must be provided as google.maps.Point object, but found: "+coordinates );
  }
  this._activeSubmapCoordinates = coordinates;
};
CustomMap.prototype.getActiveSubmapClickCoordinates = function() {
  return this._activeSubmapCoordinates;
};

CustomMap.prototype.updateAliasesForLayout = function(layoutId, jsonAliases) {
  logger.debug("Updating aliases for layout: " + layoutId);

  // init layout data
  if (this.getModel().getLayoutDataById(layoutId) === undefined || this.getModel().getLayoutDataById(layoutId) === null) {
    this.getModel().initLayoutData(layoutId);
    for (var j = 0; j < this.submaps.length; j++) {
      this.submaps[j].getModel().initLayoutData(layoutId);
    }
  }

  var aliases = JSON.parse(jsonAliases);
  for (var i = 0; i < aliases.length; i++) {
    var alias = aliases[i];
    var model = this.getSubmodelById(alias.modelId);
    if (model !== null) {
      model.getModel().updateAliasForLayout(layoutId, alias);
      model.getAliasInfoWindowById(alias.idObject).update();
    } else {
      logger.warn("Unknown model: " + alias.modelId);
    }
  }

  this.retrieveMissingAliases();
};

CustomMap.prototype.getReferenceGenome = function(type, version) {
  if (this._referenceGenome[type] === undefined) {
    this._referenceGenome[type] = [];
  }
  if (this._referenceGenome[type][version] === undefined) {
    ServerConnector.sendReferenceGenomeDetailRequest(type, version);
    this._referenceGenome[type][version] = new ReferenceGenome(null);
    return null;
  } else {
    return this._referenceGenome[type][version];
  }
};

CustomMap.prototype.updateReferenceGenome = function(type, version, jsonObj) {
  if (this._referenceGenome[type] === undefined) {
    this._referenceGenome[type] = [];
  }
  this._referenceGenome[type][version] = new ReferenceGenome(jsonObj);
  this.refreshInfoWindows();
};

CustomMap.prototype.getTopMap = function() {
  return this;
};

CustomMap.prototype.setProject = function(project) {
  this._project = project;
};
CustomMap.prototype.getProject = function() {
  return this._project;
};

CustomMap.prototype.getTopOverviewImage = function() {
  return this.getProject().getTopOverviewImage();
};

CustomMap.prototype.addControl = function(element) {
  if (this._controls[element.id] !== undefined) {
    throw new Error("Element with id "+element.id+" already added to controls");
  }
  if (ControlType[element.id]===undefined) {
    throw new Error("Unknown control type: "+element.id);
  }
  this._controls[element.id]= element;
};

CustomMap.prototype.getControl = function(type) {
  if (ControlType[type]===undefined) {
    throw new Error("Unknown control type: " + type);
  }
  return this._controls[type];
};

CustomMap.prototype.setCommentDialog = function(commentDialog) {
  this._commentDialog = commentDialog;
};

CustomMap.prototype.getCommentDialog = function() {
  return this._commentDialog;
};

CustomMap.prototype.setContextMenu = function(contextMenu) {
  this._contextMenu = contextMenu;
};

CustomMap.prototype.getContextMenu = function() {
  return this._contextMenu;
};

CustomMap.prototype.setSelectionContextMenu = function(selectionContextMenu) {
  this._selectionContextMenu = selectionContextMenu;
};

CustomMap.prototype.getSelectionContextMenu = function() {
  return this._selectionContextMenu;
};

CustomMap.prototype.openCommentDialog = function() {
  var self = this;
  return ServerConnector.getClosestElementsByCoordinates({modelId:this.getActiveSubmapId(), coordinates:this.getActiveSubmapClickCoordinates()}).then(function(elements){
    var commentDialog = self.getCommentDialog();
    if (commentDialog===undefined) {
      var div = Functions.createElement({type: "div"});
      self.getElement().appendChild(div);
      commentDialog = new CommentDialog({element:div, customMap:self});
      self.setCommentDialog(commentDialog);
    }
    self.setCommentDialog(commentDialog);
    
    return commentDialog.open(elements);
  });
};

CustomMap.prototype.fetchIdentifiedElements = function(elements, complete) {
  var modelIds = [];
  var modelElements = [];
  var i;
  for (i=0;i<elements.length;i++) {
    var element = elements[i];
    if (modelElements[element.getModelId()]===undefined) {
      modelIds.push(element.getModelId());
      modelElements[element.getModelId()]=[];
    }
    modelElements[element.getModelId()].push(element);
  } 
  
  var promises = [];
  for (i=0;i<modelIds.length;i++) {
    var modelId = modelIds[i];
    promises.push(this.getSubmodelById(modelId).getModel().getByIdentifiedElements(modelElements[modelId], complete));
  }
  return Promise.all(promises);
  
};

CustomMap.prototype.refreshSearchAutocomplete = function(){
  var self = this;
  self._searchAutocomplete=[];
  return new Promise(function(resolve, reject){
    return ServerConnector.getSuggestedQueryList().then(function(queries){
      for (var i = 0; i < queries.length; i++) {
        var mainString = queries[i];
        for (var j = 0; j < mainString.length; j++) {
          var substring = mainString.substring(0, j + 1);
          if (self._searchAutocomplete[substring] !== undefined) {
            continue;
          }

          var list = [];
          for (var k = 0; k < 5; k++) {
            if (k + i >= queries.length) {
              break;
            } else if (queries[k + i].toLowerCase().startsWith(substring.toLowerCase())) {
              list.push(queries[k + i]);
            }
          }

          self._searchAutocomplete[substring]=list;
        }
      }
      resolve(self._searchAutocomplete);
    }).catch(reject);
  });
};

CustomMap.prototype.getSearchAutocomplete = function(query){
  if (this._searchAutocomplete===undefined) {
    this.refreshSearchAutocomplete();
    return [];
  }
  
  return this._searchAutocomplete[query];
};

CustomMap.prototype.setSelectedPolygon = function(polygonData){
  this._selectedPolygon = polygonData;
};

CustomMap.prototype.getSelectedPolygon = function(){
  return this._selectedPolygon;
};

CustomMap.prototype.getSubmaps = function(){
  var submaps = this.submaps;
  if (submaps===undefined) {
    submaps=[];
  }
  return submaps;
};

CustomMap.prototype.destroy = function(){
  var self = this;
  var commentDialog = self.getCommentDialog();
  if (commentDialog!==undefined) {
    commentDialog.destroy();
  }
    
};

module.exports = CustomMap;
