"use strict";

var logger = require('../../logger');
var functions = require('../../Functions');

var AbstractInfoWindow = require('./AbstractInfoWindow');
var AliasMarker = require('../marker/AliasMarker');
var Alias = require('../data/Alias');
var LayoutAlias = require('../data/LayoutAlias');
var pileup = require('pileup');

/**
 * Class representing info window that should be opened when clicking on alias.
 */
function AliasInfoWindow(aliasParam, map, onloadFun) {
  // call super constructor
  AbstractInfoWindow.call(this, aliasParam, map);

  if (aliasParam instanceof AliasMarker) {
    this.setAlias(aliasParam.getAliasData());
    map = aliasParam.getCustomMap();
  } else {
    this.setAlias(aliasParam);
  }

  var self = this;
  var alias = this.getAlias();

  var point = new google.maps.Point(alias.x + alias.width / 2, alias.y + alias.height / 2);
  var latLng = this.customMap.fromPointToLatLng(point);
  this.googleInfowindow = new google.maps.InfoWindow({
    content : this.content,
    position : latLng
  });

  map.getTopMap().getSubmodelById(alias.getModelId()).getModel().getAliasById(alias.getId(), true).then(
      function(alias) {
        self.update(alias);
        if (onloadFun !== undefined) {
          onloadFun();
        }
      });
  this.open();
}

AliasInfoWindow.prototype = Object.create(AbstractInfoWindow.prototype);
AliasInfoWindow.prototype.constructor = AliasInfoWindow;

/**
 * Updates content of info window. The content will be automatically obtained
 * from {@link CustomMap} instance. The only optional parameter is {@link Alias}
 * data.
 * 
 * @param newAlias
 *          optional {@link Alias} data for which this {@link AliasInfoWindow}
 *          was created
 */
AliasInfoWindow.prototype.update = function(newAlias) {
  if (this.alias === undefined && newAlias !== undefined) {
    this.alias = newAlias;
  }
  this._updateContent();
};

/**
 * Creates and returns chart representing data related to alias on different
 * layouts.
 * 
 * @returns {String} html string representing chart with data related to alias
 *          on different layouts
 */
AliasInfoWindow.prototype.createChartDiv = function() {
  var result = document.createElement("div");
  for (var i = 0; i < this.layoutAliases.length; i++) {
    var rowDiv = document.createElement("div");
    if (i % 2 === 0) {
      rowDiv.className = "mapChartRowEvenDiv";
    } else {
      rowDiv.className = "mapChartRowOddDiv";
    }
    var data = this.layoutAliases[i];

    var nameDiv = document.createElement("div");
    nameDiv.className = "mapChartNameDiv";
    nameDiv.innerHTML = this.layoutNames[i];
    rowDiv.appendChild(nameDiv);

    if (data !== undefined && data !== null) {
      var value = parseFloat(data.value);
      var color = functions.overlayToColor(data);
      var leftMarginDiv = document.createElement("div");
      leftMarginDiv.innerHTML = "&nbsp;";
      leftMarginDiv.style.float = "left";
      var centerBarDiv = document.createElement("div");
      centerBarDiv.style.width = "1px";
      centerBarDiv.style.float = "left";
      centerBarDiv.style.background = "#000000";
      centerBarDiv.innerHTML = "&nbsp;";

      var rightBarDiv = document.createElement("div");
      rightBarDiv.innerHTML = "&nbsp;";
      rightBarDiv.style.float = "left";
      rightBarDiv.style.background = color;
      rightBarDiv.style.width = Math.abs(value * 100) + "px";
      var offset = 100;
      if (!isNaN(value)) {
        rightBarDiv.innerHTML = value.toFixed(2);
        if (value > 0) {
          offset = 100;
          leftMarginDiv.style.width = offset + "px";

          rightBarDiv.style.textAlign = "right";

          rowDiv.appendChild(leftMarginDiv);
          rowDiv.appendChild(centerBarDiv);
          rowDiv.appendChild(rightBarDiv);
        } else {
          offset = 100 + (value * 100);
          leftMarginDiv.style.width = offset + "px";

          rowDiv.appendChild(leftMarginDiv);
          rowDiv.appendChild(rightBarDiv);
          rowDiv.appendChild(centerBarDiv);
        }

      } else {
        offset = 100;
        leftMarginDiv.style.width = offset + "px";
        leftMarginDiv.style.background = color;
        rightBarDiv.style.width = offset + "px";
        rightBarDiv.style.background = color;
        rowDiv.appendChild(leftMarginDiv);
        rowDiv.appendChild(centerBarDiv);
        rowDiv.appendChild(rightBarDiv);
      }
    } else {
      var emptyDiv = document.createElement("div");
      emptyDiv.innerHTML = "&nbsp;";
      emptyDiv.style.float = "left";
      emptyDiv.style.width = "201px";
      rowDiv.appendChild(emptyDiv);
    }
    result.appendChild(rowDiv);
  }
  return result;
};

/**
 * Methods that creates and return DOM object with the content of the window.
 * 
 * @returns DOM object representing html code for content of the info window
 */
AliasInfoWindow.prototype.createContentDiv = function() {
  if (this.getAlias().isComplete()) {
    var result = document.createElement("div");
    var title = document.createElement("h3");
    title.innerHTML = this.alias.type + ": " + this.alias.name;
    result.appendChild(title);

    var self = this;

    var layoutsDiv = document.createElement("div");

    result.appendChild(layoutsDiv);

    self.getCustomMap().getAliasVisibleLayoutsData(this.alias.getId()).then(function(layoutAliases) {
      self.layoutAliases = layoutAliases;
      return self.getCustomMap().getVisibleLayoutNames();
    }).then(function(layoutNames) {
      self.layoutNames = layoutNames;
      var chartDiv = self.createChartDiv();
      if (chartDiv !== null) {
        layoutsDiv.appendChild(chartDiv);
      }
      var genomicDiv = self.createGenomicDiv();
      if (genomicDiv !== null) {
        layoutsDiv.appendChild(genomicDiv);
      }
    });

    return result;
  } else {
    return this.createWaitingContentDiv();
  }
};

/**
 * Returns array with data taken from all known {@link OverlayCollection}.
 * 
 * @returns array with data from {@link OverlayCollection}
 */
AliasInfoWindow.prototype.getOverlaysData = function(general) {
  return this.getCustomMap().getTopMap().getOverlayDataForAlias(this.getAlias(), general);
};

/**
 * Method returning identifier of the object for which this window was created.
 */
AliasInfoWindow.prototype.getElementId = function() {
  if (this.alias !== undefined) {
    return this.alias.getId();
  } else {
    return null;
  }
};

/**
 * Method returning type of the object for which this window was created.
 */
AliasInfoWindow.prototype.getElementType = function() {
  return "ALIAS";
};

AliasInfoWindow.prototype.createGenomicDiv = function() {
  var result = document.createElement("div");
  result.style.width = '640px';

  var titleElement = document.createElement("h3");
  titleElement.innerHTML = "Gene variants";
  result.appendChild(titleElement);

  var contentElement = document.createElement("div");
  result.appendChild(contentElement);

  var geneticInformation = false;
  var genomes = [];
  var pileupSource = [ {
    viz : pileup.viz.scale(),
    name : 'Scale'
  }, {
    viz : pileup.viz.location(),
    name : 'Location'
  } ];
  var pileupRange = {
    contig : 'chr1',
    start : 3000000000,
    stop : 0
  };

  var globalGeneVariants = [];
  var missingGeneVariantsData = false;
  for (var i = 0; i < this.layoutAliases.length; i++) {
    var data = this.layoutAliases[i];
    globalGeneVariants.push([]);

    if (data !== undefined && !missingGeneVariantsData) {
      if (data.getType() === LayoutAlias.LIGHT || data.getType() === undefined) {
        missingGeneVariantsData = true;
      } else if (data.getType() === LayoutAlias.GENETIC_VARIANT) {
        var geneVariants = data.getGeneVariants();
        for (var j = 0; j < geneVariants.length; j++) {
          if (!missingGeneVariantsData) {
            var variant = geneVariants[j];

            globalGeneVariants[i].push(variant);

            var genome = this.getCustomMap().getReferenceGenome(variant.getReferenceGenomeType(),
                variant.getReferenceGenomeVersion());

            if (variant.getContig() !== undefined) {
              pileupRange.contig = variant.getContig();
            }
            pileupRange.start = Math.min(pileupRange.start, variant.getPosition());
            var length = Math.max(variant.getModifiedDna().length, variant.getOriginalDna().length);
            pileupRange.stop = Math.max(pileupRange.stop, variant.getPosition() + length);

            if (genome !== undefined && genome.getUrl() !== undefined) {
              if (genomes[genome.getUrl()] === undefined) {

                geneticInformation = true;
                genomes[genome.getUrl()] = true;
                pileupSource.splice(0, 0, {
                  viz : pileup.viz.genome(),
                  isReference : pileupSource.length === 2,
                  data : pileup.formats.twoBit({
                    url : genome.getUrl()
                  }),
                  name : 'Reference ' + variant.getReferenceGenomeVersion()
                });
                for (var k = 0; k < genome.getGeneMappings().length; k++) {
                  var mapping = genome.getGeneMappings()[k];
                  pileupSource.push({
                    viz : pileup.viz.genes(),
                    data : pileup.formats.bigBed({
                      url : mapping.getUrl()
                    }),
                    name : 'Genes ' + mapping.getName()
                  });
                }

              }
            } else {
              logger.debug("Genome for " + variant.getReferenceGenomeType() + "," + variant.getReferenceGenomeVersion()
                  + " not loaded");
              missingGeneVariantsData = true;
            }
          }
        }
      }
    }
  }
  if (missingGeneVariantsData) {
    geneticInformation = false;
  }

  if (geneticInformation) {
    for (var iter = 0; iter < this.layoutAliases.length; iter++) {
      if (globalGeneVariants[iter].length > 0) {
        var vcfContent = this.createVcfString(globalGeneVariants[iter]);
        pileupSource.push({
          viz : pileup.viz.variants(),
          data : pileup.formats.vcf({
            content : vcfContent
          }),
          name : this.layoutNames[iter] + ' - Variants',
          options : {
            variantHeightByFrequency : true
          },
        });
      }
    }

    pileupRange.stop = Math.max(pileupRange.stop, pileupRange.start + 50);

    pileupRange.start--;
    pileupRange.stop++;

    if (this.pileup !== undefined) {
      logger.debug("Destroy pileup");
      this.pileup.destroy();
      logger.debug("Pileup destroyed");
    }
    this.pileup = pileup.create(contentElement, {
      range : pileupRange,
      tracks : pileupSource
    });

    return result;
  } else {
    return null;
  }
};

AliasInfoWindow.prototype.createVcfString = function(geneVariants) {
  var result = "";
  result += "##fileformat=VCFv4.0\n";
  result += "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n";
  for (var i = 0; i < geneVariants.length; i++) {
    var variant = geneVariants[i];
    var additionalInfo = "";
    if (variant.getAllelFrequency() !== undefined) {
      additionalInfo = "AF=" + variant.getAllelFrequency();
    }
    var variantId = ".";
    if (variant.getVariantIdentifier() !== undefined) {
      variantId = variant.getVariantIdentifier();
    }
    result += variant.getContig() + "\t" + //
    variant.getPosition() + "\t" + //
    variantId + "\t" + //
    variant.getOriginalDna() + "\t" + //
    variant.getModifiedDna() + "\t" + //
    "100.0\t" + //
    "PASS\t" + //
    additionalInfo + "\n";
  }
  return result;
};

AliasInfoWindow.prototype.setAlias = function(alias) {
  if (alias === undefined || alias === null || (!(alias instanceof Alias))) {
    throw new Error("invalid alias");
  }
  this.alias = alias;
};
AliasInfoWindow.prototype.getAlias = function() {
  return this.alias;
};

module.exports = AliasInfoWindow;
